//##############################################################################
//         Copyright (C) 2014 
//         Michele Borassi <michele.borassi@imtlucca.it>
//         David Coudert <david.coudert@inria.fr>
//         Pierluigi Crescenzi <pierluigi.crescenzi@unifi.it>
//         David Coudert <andreamarino.it@gmail.com>
//
//     Distributed under the terms of the GNU General Public License (GPL)
//                        http://www.gnu.org/licenses/
//##############################################################################

/*

  NOTE:
  - Connectivity: We assume that the input graph is connected and preferably
    bi-connected
  - The graph must be stored in the edgelist format. We assume that vertex
    labels are integers.


  BUG FIX:

  - Aurelien Lancin (Oct. 2014): fix consistency error with openmp. The bounds
    were not propagated properly among the threads and so the final result was
    sometimes incorrect.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <stdint.h>    // to use uint16_t, uint32_t, uint64_t, etc.
#include <inttypes.h>  // to use: uint64_t t; printf("%" PRIu64 "\n", t);

// To control memory usage
#define uintN_t uint32_t // already graphs with up to 4G nodes
#define uintD_t uint16_t // assume diameter at most 2^16-1. Change if needed
#define uintB_t uint8_t // boolean values


/*******************************************************************************/
/* Utility functions                                                           */
/*******************************************************************************/

#define _MAX_(_x_,_y_) ((_x_)>(_y_)?(_x_):(_y_))
#define _MIN_(_x_,_y_) ((_x_)<(_y_)?(_x_):(_y_))
#define _ITR_(_x_,_y_) ((_x_ - 1) * (_y_ - 2))
#define _ITR2_(_x_) ((_x_ - 1) * _x_)
#define MAX_LINE_LENGTH 1000


double t = 0.0;
int64_t nt = 0;
long *farness;     // array of farness of each node

// Raise an error and exit the program
void raise_error(char *message){
    fprintf(stderr,"%s\n",message);
    exit(1);
}

// Raise an error if ptr is NULL
void test_ptr(void *ptr, char *message){
    if (ptr == NULL )
        raise_error(message);
}

// Returns current time in seconds
float getTimeSeconds() {
    float time = clock();
    return time / 1000000;
}


/*******************************************************************************/
/* Pairs tools                                                                 */
/*******************************************************************************/

// type pair_t: pair of vertices with distance
typedef struct _pair_t_ {
    uintN_t x;
    uintN_t y;
    uintD_t d;
    struct _pair_t_ *next; // to make lists
} pair_t;

pair_t* pair_new(uintN_t x, uintN_t y, uintD_t value, pair_t *next){
    pair_t *p;
    test_ptr( (p=(pair_t*)malloc(sizeof(pair_t))), "pair_new: malloc() error");
    p->x = _MIN_(x,y);
    p->y = _MAX_(x,y);
    p->d = value;
    p->next = next;
    return p;
}

void pair_free_list(pair_t *p){
    pair_t *q;
    while(p) {
        q = p;
        p = p->next;
        free(q);
    }
}
// Comparison function for qsort().
// We want decreasing order of distances and increasing lexicographic order when
// same distance
static int pairs_compare_distance_new(void const *a, void const *b){
    int res;
    struct _pair_t_ const *pa = a;
    struct _pair_t_ const *pb = b;

    res = pb->d - pa->d;
    if (res == 0)
    {
        res = farness[pa->x] + farness[pa->y] - farness[pb->x] - farness[pb->y];
        //        res = pa->x - pb->x;
        //        if (res == 0)
        //        {
        //            res = pa->y - pb->y;
        //        }
    }
    return res;
}

// Comparison function for qsort().
// We want decreasing order of distances and increasing lexicographic order when
// same distance
static int pairs_compare_distance(void const *a, void const *b){
    int res;
    struct _pair_t_ const *pa = a;
    struct _pair_t_ const *pb = b;

    res = pb->d - pa->d;
    if (res == 0)
    {
        res = pa->x - pb->x;
        if (res == 0)
        {
            res = pa->y - pb->y;
        }
    }
    return res;
}

// sort pairs by decreasing distances
// void pair_array_sort_distance(pair_t *p, uint64_t nb_pairs){
//   qsort(p, nb_pairs, sizeof(pair_t), pairs_compare_distance);
// }
#define pair_array_sort_distance(_p_, _nb_pairs_) qsort(_p_, _nb_pairs_, sizeof(pair_t), pairs_compare_distance_new)

/*******************************************************************************/
/* List structures                                                             */
/*******************************************************************************/

typedef struct node {
    int index;
    struct node* next;
} node;

/*******************************************************************************/
/* Graph tools                                                                 */
/*******************************************************************************/

typedef struct _node_t_ {
    uintN_t id;             // index of the node in [0..N-1]
    uintN_t true_id;        // id of the node in input file, assumed < 2^sizeof(intN_t)
    struct _node_t_ *next;  // to make a list
} node_t;

typedef struct _edge_t_ {
    uintN_t u;              // end node with smallest index
    uintN_t v;              // end node with largest index
    struct _edge_t_ *next;
} edge_t;

typedef struct _graph_t_ {
    uintN_t n;                    // number of nodes
    uintN_t m;                    // number of edges
    node_t *nodes;                // list of nodes
    edge_t *edges;                // list of edges
    uintN_t *degree;              // degree of the vertices. Array index by nodes id
    edge_t ***incident_edges;     // arrays of incident edges of nodes
    node_t **int_to_nodes;        // mapping node id to node
    uintN_t *true_id_to_int;      // mapping id in input file to ids in [0..n-1]
    uintN_t _size_true_id_to_int; // largest index in input file
} graph_t;



graph_t* graph_new(){
    graph_t *g;

    test_ptr( (g = (graph_t*)malloc(sizeof(graph_t))), "graph_new: malloc() error");
    g->n = 0;
    g->m = 0;
    g->nodes = NULL;
    g->edges = NULL;
    g->degree = NULL;
    g->int_to_nodes = NULL;
    g->true_id_to_int = NULL;
    g->_size_true_id_to_int = 0;
    g->incident_edges = NULL;
    return(g);
}


void graph_free(graph_t *g){
    edge_t *e;
    node_t *u;
    uintN_t i;
    if (g!=NULL) {
        while(g->edges!=NULL) {
            e = g->edges;
            g->edges = e->next;
            free(e);
        }
        while(g->nodes!=NULL) {
            u = g->nodes;
            g->nodes = u->next;
            free(u);
        }
        if (g->int_to_nodes!=NULL)
            free(g->int_to_nodes);
        if(g->true_id_to_int!=NULL)
            free(g->true_id_to_int);
        if (g->incident_edges!=NULL){
            for(i=0;i<g->n;i++)
                if (g->incident_edges[i])
                    free(g->incident_edges[i]);
            free(g->incident_edges);
        }
        if (g->degree)
            free(g->degree);
        free(g);
    }
}


// add node to graph if not already in
void _graph_add_node_(graph_t *g, uintN_t u){
    node_t *v;

    if(g->true_id_to_int[u]==(uintN_t)-1){
        test_ptr( (v = (node_t*)malloc(sizeof(node_t))), "_graph_add_node_: malloc() error");
        v->true_id = u;
        v->id = u;
        g->true_id_to_int[u] = v->id;
        g->n += 1;
        v->next = g->nodes;
        g->nodes = v;
    }
}



// return true if the graph has edge u,v
int graph_has_edge(graph_t *g, uintN_t u, uintN_t v) {
    edge_t *e;
    uintN_t a = _MIN_(u,v);
    uintN_t b = _MAX_(u,v);

    e = g->edges;
    while(e){
        if( (e->u==a)&&(e->v==b) )
            return 1;
        e = e->next;
    }
    return 0;
}

// add edge to the graph. Multiple edges are allowed here
void graph_add_edge(graph_t *g, uintN_t u, uintN_t v){
    edge_t *e;

    test_ptr( (e = (edge_t*)malloc(sizeof(edge_t))), "graph_add_edge: malloc() error");
    e->u = _MIN_(u,v);
    e->v = _MAX_(u,v);
    e->next = g->edges;
    g->edges = e;
    g->m += 1;
}


// consolidate the graph from lists of edges
graph_t* graph_consolidate(pair_t *p){
    edge_t *e;
    node_t *v;
    uintN_t i, *tmp;
    uintN_t true_id_max = 0;
    pair_t *q;
    graph_t *g = graph_new();

    // We create a list of nodes and mappings
    // we first search for the largest index
    q = p;
    while(q){
        true_id_max = _MAX_(true_id_max, q->y);
        q = q->next;
    }
    // we prepare mapping true_id_to_int
    test_ptr( (g->true_id_to_int = (uintN_t*)malloc(sizeof(uintN_t)*(true_id_max+1))), "graph_create_mapping: malloc() error");
    memset(g->true_id_to_int, (char)-1, sizeof(uintN_t)*(true_id_max+1));
    g->_size_true_id_to_int = true_id_max+1;
    // we create the list of nodes
    q = p;
    while(q) {
        _graph_add_node_(g, q->x);
        _graph_add_node_(g, q->y);
        q = q->next;
    }
    // we now make the mapping int_to_nodes
    test_ptr( (g->int_to_nodes = (node_t**)malloc(sizeof(node_t*)*g->n)), "graph_create_mapping: malloc() error");
    v = g->nodes;
    while(v){
        g->int_to_nodes[v->id] = v;
        v = v->next;
    }

    // We build set of edges with end nodes in 0..N-1
    q = p;
    while(q) {
        graph_add_edge(g, g->true_id_to_int[q->x], g->true_id_to_int[q->y]);
        q = q-> next;
    }

    // We compute degree of nodes
    test_ptr( (g->degree=(uintN_t*)malloc(sizeof(uintN_t)*g->n)), "graph_consolidate: malloc() error");
    for(i=0;i<g->n;i++)
        g->degree[i] = 0;
    e = g->edges;
    while(e) {
        g->degree[e->u] += 1;
        g->degree[e->v] += 1;
        e = e->next;
    }

    // We construct list of incident edges
    test_ptr( (tmp = (uintN_t*)malloc(sizeof(uintN_t)*g->n)), "graph_consolidate: malloc() error");
    memset(tmp, (char)0, sizeof(uintN_t)*g->n);
    test_ptr( (g->incident_edges = (edge_t***)malloc(sizeof(edge_t**)*g->n)), "graph_consolidate: malloc() error");
    for(i=0;i<g->n;i++)
        test_ptr((g->incident_edges[i] = (edge_t**)malloc(sizeof(edge_t*)*g->degree[i])), "graph_consolidate: malloc() error");

    e = g->edges;
    while(e) {
        g->incident_edges[e->u][tmp[e->u]] = e;
        g->incident_edges[e->v][tmp[e->v]] = e;
        tmp[e->u] += 1;
        tmp[e->v] += 1;
        e = e->next;
    }

    free(tmp);
    return g;
}


// Read edgelist format
graph_t *graph_read_edgelist(FILE *f, char verbose){
    char line[MAX_LINE_LENGTH], *ll;
    uintN_t u, v;
    pair_t *p = NULL;
    graph_t *g;

    if(verbose) fprintf(stderr,"read graph\n");
    // read data from file
    while (fgets(line, MAX_LINE_LENGTH, f)){
        // remove useless spaces or tabs at the beginning of the line
        ll = line;
        while((ll[0]==' ')||(ll[0]=='\t')) ll = ll+1;

        switch (ll[0]){
        case '#': // read comments and do nothing
            break;
        case 'c': // read comments and do nothing
            break;
        case '\n': // read empty line and do nothing
            break;
        default: // read edge {u,v}
            if (sscanf(ll, "%u %u\n", &u, &v) != 2)    // try with space delimiter
                if (sscanf(ll, "%u\t%u\n", &u, &v) != 2) // try with tab delimiter
                {fprintf(stderr,"Last line: %s",line);
                    raise_error("graph_read_edgelist: sscanf error while reading <u> <v>");}
            if(u!=v)
                p = pair_new(_MIN_(u,v), _MAX_(u,v), 0, p);
        }
    }

    g = graph_consolidate(p);

    pair_free_list(p);
    return g;
}





/*******************************************************************************/
/* Distances                                                                   */
/*******************************************************************************/

// Added array ecc to copmpute eccentricities of nodes
void distances_and_far_apart_pairs(graph_t *g, uintD_t **distances, uintD_t *ecc, char **far_apart_pairs, char verbose){
    uintN_t *seen;
    uintN_t *waiting_list;
    uintN_t waiting_beginning, waiting_end;
    uintN_t source, u, v, i;
    uintD_t *distance_s;

    if(!g) raise_error("distances_and_far_apart_pairs: input graph is NULL");
    if(g->n==0) raise_error("distances_and_far_apart_pairs: empty graph");

    test_ptr( (seen=(uintN_t*)malloc(sizeof(uintN_t)*g->n)), "distances_and_far_apart_pairs: malloc() error");
    test_ptr( (waiting_list=(uintN_t*)malloc(sizeof(uintN_t)*g->n)), "distances_and_far_apart_pairs: malloc() error");

    memset(seen, (char)-1, sizeof(uintN_t)*g->n);

    for(i=0;i<g->n;i++) {
        memset(far_apart_pairs[i], (char)-1, sizeof(char)*g->n);
        far_apart_pairs[i][i] = 0;
    }

    for(source=0;source<g->n;source++) {

        // the source is seen
        seen[source] = source;

        // Its parameters can already be set
        distance_s = distances[source];
        memset(distance_s, (char)-1, sizeof(uintD_t)*g->n);
        distance_s[source] = 0;

        // and added to the queue
        waiting_list[0] = source;
        waiting_beginning = 0;
        waiting_end = 0;

        // For as long as there are vertices left to explore
        while(waiting_beginning <= waiting_end) {

            // We pick the first one
            v = waiting_list[waiting_beginning];

            // Iterating over all the outneighbors u of v
            for(i=0;i<g->degree[v];i++) {
                u = g->incident_edges[v][i]->u;
                if(u==v) u = g->incident_edges[v][i]->v;

                // If we notice one of these neighbors is not seen yet, we set
                // its parameters and add it to the queue to be explored later.
                if (seen[u]!=source) {
                    distance_s[u] = distance_s[v]+1;
                    // v is on the path from source to u
                    far_apart_pairs[source][v] = 0;
                    far_apart_pairs[v][source] = 0;
                    seen[u] = source;
                    waiting_end += 1;
                    waiting_list[waiting_end] = u;
                }
                else if(distance_s[u] == distance_s[v]+1) {
                    // v is on the path from source to u
                    far_apart_pairs[source][v] = 0;
                    far_apart_pairs[v][source] = 0;
                }
            }
            waiting_beginning += 1;
        }
        ecc[source] = distance_s[waiting_list[g->n-1]];
    }
    free(seen);
    free(waiting_list);
}


// Compute Distances from source using BFS
void single_source_distances_BFS(graph_t *g, uintN_t source, uintN_t *distance_s,
                                 char *seen, uintN_t *waiting_list) {
    uintN_t waiting_beginning, waiting_end;
    uintN_t u, v, i;

    memset(seen, 0, g->n);
    memset(distance_s, (char)-1, sizeof(uintN_t)*g->n);

    // the source is seen
    seen[source] = 1;
    distance_s[source] = 0;

    // and added to the queue
    waiting_list[0] = source;
    waiting_beginning = 0;
    waiting_end = 0;

    // For as long as there are vertices left to explore
    while(waiting_beginning <= waiting_end) {

        // We pick the first one
        v = waiting_list[waiting_beginning];

        // Iterating over all the outneighbors u of v
        for(i=0;i<g->degree[v];i++) {
            u = g->incident_edges[v][i]->u;
            if(u==v) u = g->incident_edges[v][i]->v;

            // If we notice one of these neighbors is not seen yet, we set
            // its parameters and add it to the queue to be explored later.
            if (!seen[u]) {
                distance_s[u] = distance_s[v]+1;
                // v is on the path source to u
                seen[u] = 1;
                waiting_end += 1;
                waiting_list[waiting_end] = u;
            }
        }
        waiting_beginning += 1;
    }
}

// Return 1 if the graph is connected and 0 otherwise
int graph_is_connected(graph_t *g){
    uintN_t *waiting_list, *distances;
    char *seen;
    uintN_t i,res;

    test_ptr( (distances = (uintN_t*)malloc(g->n*sizeof(uintN_t))), "graph_is_connected: malloc error");
    test_ptr( (waiting_list = (uintN_t*)malloc(g->n*sizeof(uintN_t))), "graph_is_connected: malloc error");
    test_ptr( (seen = (char*)malloc(g->n*sizeof(char))), "graph_is_connected: malloc error");

    single_source_distances_BFS(g, 0, distances, seen, waiting_list);
    for(i=0,res=1;i<g->n;i++)
        if(distances[i]>g->n){
            res = 0;
            break;
        }
    free(seen);
    free(distances);
    free(waiting_list);
    return res;
}


/*******************************************************************************/
/* Data for hyperbolicity algorithms                                           */
/*******************************************************************************/
// Added array ecc to store eccentricities of nodes

// type data_t: all necessary data
typedef struct _data_t_ {
    uintN_t n;         // number of nodes
    uint64_t nb_pairs; // number of distant pairs <= n(n-1)/2
    pair_t *pair;      // array of nb_pairs pairs
    uintD_t **dist;    // distance matrix
    uintD_t *ecc;      // array of eccentricities
} data_t;

data_t* data_new(uintN_t n, uint64_t nb_pairs, uintD_t **distances, uintD_t *ecc) {
    data_t *data;
    test_ptr((data = (data_t*) malloc(sizeof(data_t))), "data_new: malloc error");
    data->n = n;
    data->nb_pairs = nb_pairs;
    test_ptr((data->pair = (pair_t*) malloc(sizeof(pair_t)*nb_pairs)), "data_new: malloc error");
    // This is the bug I mentioned in the email.
    data->dist = distances;
    data->ecc = ecc;
    return data;
}

void data_free(data_t *data) {
    free(data->pair);
    free(data->dist[0]);
    free(data->dist);
    free(data->ecc);
    free(data);
}

data_t* data_make(graph_t *g, int verbose) {
    data_t *data;
    uintN_t i,j;
    uint64_t k;
    uintD_t **distances;
    uintD_t *ecc;
    char **far_apart_pairs;

    test_ptr( (distances = (uintD_t**)malloc(g->n*sizeof(uintD_t*))), "data_make: malloc error");
    test_ptr( (ecc = (uintD_t*)malloc(g->n*sizeof(uintD_t))), "data_make: malloc error");
    test_ptr( (farness = (long*)malloc(g->n*sizeof(long))), "data_make: malloc error");
    test_ptr( (distances[0] = (uintD_t*)malloc(g->n*g->n*sizeof(uintD_t))), "data_make: malloc error");
    test_ptr( (far_apart_pairs = (char**)malloc(g->n*sizeof(char*))), "data_make: malloc error");
    test_ptr( (far_apart_pairs[0] = (char*)malloc(g->n*g->n*sizeof(char))), "data_make: malloc error");
    for(i = 1; i < g->n; i++) {
        distances[i] = distances[i-1] + g->n;
        far_apart_pairs[i] = far_apart_pairs[i-1]+g->n;
    }

    distances_and_far_apart_pairs(g, distances, ecc, far_apart_pairs, verbose);

    for(i = 0,k = 0;i<g->n;i++)
        for(j = i+1;j<g->n;j++) {
            if(far_apart_pairs[i][j]) {
                k += 1;
            }
        }

    for (i = 0; i < g->n; i++) {
        uintD_t *disti = distances[i];
        farness[i] = 0;
        for (j = 0; j < g->n; j++) {
            farness[i] += disti[j];
        }
    }

    data = data_new(g->n, k, distances, ecc);

    for(i=0,k=0;i<g->n;i++)
        for(j=i+1;j<g->n;j++)
            if(far_apart_pairs[i][j]) {
                data->pair[k].x = i;
                data->pair[k].y = j;
                data->pair[k].d = data->dist[i][j];
                k += 1;
            }

    pair_array_sort_distance(data->pair, data->nb_pairs);

    free(far_apart_pairs[0]);
    free(far_apart_pairs);
    return data;
}


/*******************************************************************************/
/* Hyperbolicity                                                               */
/*******************************************************************************/

// This method implements the counting sort algorithm as described http://en.wikipedia.org/wiki/Counting_sort
// We assume that all keys are in 0..N-1
uintN_t *counting_sort_new(uintD_t *key, uintN_t n, int revert) {
    uintN_t * count;
    uintN_t *output;
    uintN_t kmin, kmax, total, oldcount;
    int i;

    test_ptr( (count = (uintN_t*) malloc(n * sizeof(uintN_t))), "counting_sort: malloc error");
    test_ptr( (output = (uintN_t*) malloc(n * sizeof(uintN_t))), "counting_sort: malloc error");
    memset(count, 0, sizeof(uintN_t)*n);
    
    // calculate the histogram of key frequencies:
    kmax = 0;
    for(i=0;i<n;i++){
        count[key[i]]++;
        kmax = _MAX_(kmax, key[i]);
    }
    
    // calculate the starting index for each key:
    if(revert==1){
        total = 0;
        for(i=kmax;i>=0;i--){
            oldcount = count[i];
            count[i] = total;
            total += oldcount;
        }
    }
    else {
        total = 0;
        for(i=0;i<=kmax;i++){
            oldcount = count[i];
            count[i] = total;
            total   += oldcount;
        }
    }

    // copy to output array, preserving order of inputs with equal keys:
    for(i=0;i<n;i++){
        output[count[key[i]]] = i;
        count[key[i]]++;
    }
    
    free(count);
    
    return output;
}


float hyperbolicity_old(data_t *data, graph_t *g) {
#ifdef COUNT
    int64_t cpt = 0;
#endif

    pair_t *pair = data->pair;
    pair_t pi, pj;
    uint16_t **dist = data->dist;
    uintD_t *distx, *disty;
    uintD_t delta = 0, UB = (uintD_t) -1;
    int32_t S2, S3, tmp, delta_dxy, z, r, y;
    int64_t i = 0, j = 0, nb_pairs, last_i = 0, retains = 0;
    nt = 0;

    nb_pairs = data->nb_pairs;

    for (i = 1; i < nb_pairs; i++) {
        pi = *(pair+i);

        if (pi.d < UB) {
            UB = pi.d;
        }

        if (delta >= pi.d) {

            // weird observation, using i = nb_pairs as end condition for the first
            // loop and break for the second is sometimes more efficient than using
            // two breaks O_o
            last_i = i;
            i = nb_pairs;
            //break;
        }
        else {
            y = pi.y;
            distx = dist[pi.x];
            disty = dist[y];
            delta_dxy = delta - distx[y];

            for (j = 0; j < i; j++) {

                nt++;
                pj = *(pair+j);
                z = pj.x;
                r = pj.y;
                S2 = distx[z] + disty[r];
                S3 = distx[r] + disty[z];
                tmp = pj.d - _MAX_( S2, S3 );

                if (tmp > delta_dxy) {
                    delta = tmp + distx[y];


                    if (delta >= pj.d) {
                        retains = i - j - 1;
                        break;
                    }
                    else {
                        delta_dxy = tmp;
                    }
                }
            }
        }
    }

    return (delta / 2.0);
}



// New version of the hyperbolicity algorithm.
float hyperbolicity_new(data_t *data, graph_t *g) {
    pair_t *pair = data->pair;
    pair_t pi;
    uintD_t **dist = data->dist;
    uintD_t *distx, *disty, *distc;
    uintB_t *acc;
    uintN_t *acc_acc;
    uintD_t UB = (uintD_t) -1;
    int16_t S2, S3, delta = 0;
    uintN_t x, y, c, v, w, j;
    int64_t i = 0, tmp, nb_pairs, last_i = 0, q;
    nt = 0;
    uintN_t N = g->n;
    uintN_t *nodesToConsider;
    uintD_t *nodesToConsiderBoolean;
    int cont_nodesToConsider = 0;

    // Allocate memory for nodesToConsider.
    test_ptr( (nodesToConsider = (uintN_t*) malloc(N * sizeof(uintN_t))), "hyperbolicity_single_thread: malloc error");
    test_ptr( (nodesToConsiderBoolean = (uintD_t*) calloc(N, sizeof(uintD_t))), "hyperbolicity_single_thread: malloc error");


    // Allocate memory for acc(eptable) and value arrays.
    test_ptr( (acc = (uintB_t*) calloc(N, sizeof(uintB_t))), "hyperbolicity_single_thread: malloc error");
    test_ptr( (acc_acc = (uint32_t*) malloc(N * sizeof(uint32_t))), "hyperbolicity_single_thread: malloc error");


    // Definition and initialization of array of mate lists. Later, mate[v] will be a list of all
    // vertices w such that pair (v,w) has already been analyzed.
    uintN_t **mate;
    uintN_t *cont_mate;

    test_ptr((mate = (uintN_t**) malloc(N * sizeof(uintN_t*))), "hyperbolicity_single_thread: malloc error");
    test_ptr((cont_mate = (uintN_t*) calloc(N, sizeof(uintN_t))), "hyperbolicity_single_thread: malloc error");


    for (v = 0; v < N; v++) {
        test_ptr((mate[v] = (uintN_t*) malloc(N * sizeof(uintN_t))), "hyperbolicity_single_thread: malloc error");
    }


    // Set c as central node (node with smallest farness), that will be used later in the algorithm.
    // TODO: try other choices.
    c = 0;
    tmp = farness[0];
    for (i = 0; i < N; i++)
        if (farness[i] < tmp) {
            c = i;
            tmp = farness[i];
        }
    distc = dist[c];

    // Scan all not far apart pairs (x,y)
    nb_pairs = data->nb_pairs;
    int nPairs = 0;

    for (i = 0; i < nb_pairs; i++) {

        pi = *(pair+i);
        nPairs++;
        if (pi.d < UB) {
            UB = pi.d;
        }
        if (delta >= pi.d) {
            // Hyperbolicity has been computed since all other pairs have smaller distance
            last_i = i;
            i = nb_pairs;
        } else {
            // This is the new part.

            y = pi.y;
            x = pi.x;

            // Update nodesToConsider.
            if (!nodesToConsiderBoolean[x]) {
                nodesToConsider[cont_nodesToConsider++] = x;
                nodesToConsiderBoolean[x] = 1;
            }
            if (!nodesToConsiderBoolean[y]) {
                nodesToConsider[cont_nodesToConsider++] = y;
                nodesToConsiderBoolean[y] = 1;
            }
            // Update mate variable.
            mate[x][cont_mate[x]++] = y;
            mate[y][cont_mate[y]++] = x;

            distx = dist[x];
            disty = dist[y];


            // Fills the arrays of acceptable and valuable nodes. When analyzing pair (x,y),
            // a node is acceptable if all the following conditions hold: if one of them does not hold,
            // then, for each w, delta(x,y,v,w) <= delta. A node is valuable if a further condition
            // holds, such that, if delta(x,y,v,w) > delta, then either v or w must satisfy that condition.

            // Number of acceptable nodes.
            int na = 0;
            // Number of valuable nodes.
            int naa = 0;
            uintD_t deltaplusone = (uint16_t) delta + 1;
            uintD_t *ecc = data->ecc;

            // These variables are used to speed-up the computation later.
            uintD_t condacc1, condacc2, condval = distx[y] - delta;

            memset(acc, 0, N);

            if (3 * deltaplusone >= 2 * distx[y]) {
                // Michele: I do not like this, but I don't know how to improve it
                // (conversion problems with the sign of condacc1 and condacc2).
                condacc1 = 3 * deltaplusone - 2 * distx[y];
                condacc2 = 2 * deltaplusone - distx[y];
                for (j = cont_nodesToConsider; j--;) {
                    v = nodesToConsider[j];
                    // Test of acceptable vertices.

                    if (2 * (ecc[v] - _MAX_(distx[v], disty[v])) >= condacc1) {
                        if(2 * distx[v] >= deltaplusone && 2 * disty[v] >= deltaplusone) {
                            if (2 * data->ecc[v] >= condacc2 + distx[v] + disty[v]) {
                                acc[v] = 1;
                                na++;
                                // Test for valuable vertices
                                if (2 * distc[v] + condval > distx[v] + disty[v]) {
                                    acc_acc[naa++] = v;
                                }
                            }
                        }
                    }
                }
            } else {
                for (j = cont_nodesToConsider; j--;) {
                    v = nodesToConsider[j];

                    if(2 * distx[v] >= deltaplusone && 2 * disty[v] >= deltaplusone) {
                        if (2 * data->ecc[v] + distx[y] >= 2 * deltaplusone + distx[v] + disty[v] ) {
                            acc[v] = 1;
                            na++;
                            // Test for valuable vertices
                            if (2 * distc[v] + condval > distx[v] + disty[v]) {
                                acc_acc[naa++] = v;
                            }
                        }
                    }
                }
            }


            // Analyze 4-tuples of the form (x,y,v,w) with v valuable, and w such that
            // (v,w) < (x,y) in the original sorting of pairs (that is, w belongs to mate[v]).
            for (q = naa; q--; ) {
                v = acc_acc[q];
                for (j = cont_mate[v]; j--;) {
                    w = mate[v][j];
                    if (acc[w]) {
                        nt++;
                        S2 = distx[v] + disty[w];
                        S3 = distx[w] + disty[v];
                        delta = _MAX_((int16_t)(distx[y] + dist[v][w]) - _MAX_( S2, S3 ), delta);
                    }
                }
            }
        }
    }

    // Free mate list.
    for (v = 0; v < N; v++) {
        free(mate[v]);
    }
    free(mate);
    free(cont_mate);
    free(nodesToConsider);
    free(nodesToConsiderBoolean);
    free(acc);
    free(acc_acc);
    return (delta / 2.0);
}




// New version of the hyperbolicity algorithm.
float hyperbolicity_new_improved(data_t *data, graph_t *g) {
    pair_t *pair = data->pair;
    pair_t pi;
    uint16_t **dist = data->dist;
    uintD_t *distx, *disty, *distc;
    uintB_t *acc;
    uint32_t *acc_acc;
    uintD_t UB = (uintD_t) -1;
    uintN_t x, y, c, v, w, j;
    int16_t S2, S3, delta = 0;
    int64_t i = 0, tmp, nb_pairs, last_i = 0, q;
    nt = 0;
    uintN_t N = g->n;
    uintD_t *nodesToConsiderBoolean;
    uintN_t **nodesToConsider;
    uintD_t *valuesCompNodesToConsider;
    uintD_t *ecc = data->ecc;

    t -= getTimeSeconds(); // Used to measure time of subroutines, printed at the end.

    // Allocate memory for nodesToConsider.
    test_ptr( (nodesToConsiderBoolean = (uintD_t*) calloc(N, sizeof(uintD_t))), "hyperbolicity_single_thread: malloc error");
    test_ptr( (nodesToConsider = (uintN_t**) malloc(N * sizeof(uintN_t*))), "hyperbolicity_single_thread: malloc error");
    test_ptr( (valuesCompNodesToConsider = (uintD_t*) malloc(N * sizeof(uintD_t))), "hyperbolicity_single_thread: malloc error");

    for (i = 0; i < N; i++) {
        uintD_t* disti = data->dist[i];
        for (j = 0; j < N; j++) {
            valuesCompNodesToConsider[j] = ecc[j] - disti[j];
        }
        nodesToConsider[i] = counting_sort_new(valuesCompNodesToConsider, N, 1);
    }
    free(valuesCompNodesToConsider);
    t += getTimeSeconds(); // Used to measure time of subroutines, printed at the end.

    // Allocate memory for acc(eptable) and value arrays.
    test_ptr( (acc = (uintB_t*) malloc(N * sizeof(uintB_t))), "hyperbolicity_single_thread: malloc error");
    test_ptr( (acc_acc = (uintN_t*) malloc(N * sizeof(uintN_t))), "hyperbolicity_single_thread: malloc error");

    // Definition and initialization of array of mate lists. Later, mate[v] will be a list of all
    // vertices w such that pair (v,w) has already been analyzed.
    uintN_t **mate;
    uintN_t *cont_mate;

    test_ptr((mate = (uintN_t**) malloc(N * sizeof(uintN_t*))), "hyperbolicity_single_thread: malloc error");
    test_ptr((cont_mate = (uintN_t*) calloc(N, sizeof(uintN_t))), "hyperbolicity_single_thread: malloc error");

    for (v = 0; v < N; v++) {
        test_ptr((mate[v] = (uintN_t*) malloc(N * sizeof(uintN_t))), "hyperbolicity_single_thread: malloc error");
    }

    // Set c as central node (node with smallest farness), that will be used later in the algorithm.
    // TODO: try other choices.
    c = 0;
    tmp = farness[0];
    for (i = 0; i < N; i++)
        if (farness[i] < tmp) {
            c = i;
            tmp = farness[i];
        }
    distc = dist[c];


    // Scan all not far apart pairs (x,y)
    nb_pairs = data->nb_pairs;
    int nPairs = 0;

    for (i = 0; i < nb_pairs; i++) {

        pi = *(pair+i);
        nPairs++;
        if (pi.d < UB) {
            UB = pi.d;
        }
        if (delta >= pi.d) {
            // Hyperbolicity has been computed since all other pairs have smaller distance
            last_i = i;
            i = nb_pairs;
        } else {
            // This is the new part.
            y = pi.y;
            x = pi.x;

            // Update nodesToConsider.
            nodesToConsiderBoolean[x] = 1; // should be as fast as a test.
            nodesToConsiderBoolean[y] = 1;
            
            
            
            // Update mate variable.
            mate[x][cont_mate[x]++] = y;
            mate[y][cont_mate[y]++] = x;

            distx = dist[x];
            disty = dist[y];


            // Fills the arrays of acceptable and valuable nodes. When analyzing pair (x,y),
            // a node is acceptable if all the following conditions hold: if one of them does not hold,
            // then, for each w, delta(x,y,v,w) <= delta. A node is valuable if a further condition
            // holds, such that, if delta(x,y,v,w) > delta, then either v or w must satisfy that condition.

            // Number of acceptable nodes.
            int na = 0;
            // Number of valuable nodes.
            int naa = 0;
            uintD_t deltaplusone = (uint16_t) delta + 1;
            uintD_t *ecc = data->ecc;

            // These variables are used to speed-up the computation later.
            uintD_t condacc1, condacc2, condval = distx[y] - delta;

            memset(acc, 0, N);
            if (farness[x] < farness[y]) {
                uint16_t *distxcopy = distx;
                distx = disty;
                disty = distxcopy;
                uintN_t xcopy = x;
                x = y;
                y = xcopy;
                //       printf("%d CHANGE %d\n", distx[y], disty[x]);
            }

            if (3 * deltaplusone >= 2 * distx[y]) {
                // Michele: I do not like this, but I don't know how to improve it
                // (conversion problems with the sign of condacc1 and condacc2).
                condacc1 = 3 * deltaplusone - 2 * distx[y];
                condacc2 = 2 * deltaplusone - distx[y];
                for (j = 0; j < N; j++) {
                    v = nodesToConsider[x][j];
                    // Test of acceptable vertices.

                    if (nodesToConsiderBoolean[v]) {
                        if (2 * (ecc[v] - distx[v]) >= condacc1) {
                            if (2 * (ecc[v] - disty[v]) >= condacc1) {
                                if(2 * distx[v] >= deltaplusone && 2 * disty[v] >= deltaplusone) {
                                    if (2 * ecc[v] >= condacc2 + distx[v] + disty[v] ) {
                                        acc[v] = 1;
                                        na++;
                                        // Test for valuable vertices
                                        if (2 * distc[v] + condval > distx[v] + disty[v]) {
                                            acc_acc[naa++] = v;
                                        }
                                    }
                                }
                            }
                        } else { // Improvement in this version, to consider a smaller number of vertices
                            break;
                        }
                    }
                }
            } else {
                for (v = 0; v < N; v++) {
                    if(2 * distx[v] >= deltaplusone && 2 * disty[v] >= deltaplusone) {
                        if (2 * data->ecc[v] + distx[y] >= 2 * deltaplusone + distx[v] + disty[v] ) {
                            acc[v] = 1;
                            na++;
                            // Test for valuable vertices
                            if (2 * distc[v] + condval > distx[v] + disty[v]) {
                                acc_acc[naa++] = v;
                            }
                        }
                    }
                }
            }


            // Analyze 4-tuples of the form (x,y,v,w) with v valuable, and w such that
            // (v,w) < (x,y) in the original sorting of pairs (that is, w belongs to mate[v]).
            for (q = naa; q--; ) {
                v = acc_acc[q];
                for (j = cont_mate[v]; j--;) {
                    w = mate[v][j];
                    if (acc[w]) {
                        nt++;
                        S2 = distx[v] + disty[w];
                        S3 = distx[w] + disty[v];
                        delta = _MAX_((int16_t) (distx[y] + dist[v][w]) - _MAX_(S2, S3), delta);
                    }
                }
            }
        }
    }

    for (v = 0; v < N; v++) {
        free(mate[v]);
        free(nodesToConsider[v]);
    }
    free(mate);
    free(cont_mate);
    free(nodesToConsider);
    free(nodesToConsiderBoolean);
    free(acc);
    free(acc_acc);
    return (delta / 2.0);
}


/*******************************************************************************/
/* Main functions                                                              */
/*******************************************************************************/




// Parse input parameters
void parse_input_parameters(int argc, char **argv, char *filename){
    int i;

    if (argc > 1) {
        for(i=1;i<argc;i++){
            if(!(sscanf(argv[i], "file=%s", filename) == 1))
                break;
        }
    }
}



int main(int argc, char **argv)
{
    graph_t *g;
    data_t *data = NULL;
    float delta = 0, deltacheck;
    double begin, end;
    FILE *f;             // input file
    char verbose = 0;    // to display some information during processing
    char filename[55]; // Input file name
    int64_t oldnt;

    begin = getTimeSeconds();

    FILE *out = fopen("output.txt", "a");


    filename[0]=-1;

    parse_input_parameters(argc, argv, filename);
    // Read the graph
    test_ptr((f = fopen(filename, "r")), "Error when reading input file");
    g = graph_read_edgelist(f, verbose);
    fclose(f);

    fprintf(stdout, "                              File     NN         NE    Preproc  Delta       Tuples hyp  Time hyp2  Prep hyp2  Time hyp1       Tuples CCL   Time CCL\n");
    fprintf(out, "                              File     NN         NE    Preproc  Delta       Tuples hyp  Time hyp2  Prep hyp2  Time hyp1       Tuples CCL   Time CCL\n");

    fprintf(stdout, "%34s", filename);
    fprintf(out, "%34s", filename);
    fflush(stdout);
    fflush(out);


    // Raise error if the input graph is not connected
    if(!graph_is_connected(g))
        raise_error("The input graph is not connected");

    fprintf(stdout,"%7d %10d",g->n, g->m);
    fprintf(out,"%7d %10d",g->n, g->m);

    data = data_make(g, verbose);
    end = getTimeSeconds();
    fprintf(stdout, "%11.4f", end - begin);
    fprintf(out, "%11.4f", end - begin);
    fflush(stdout);
    fflush(out);


    begin = getTimeSeconds();
    delta = hyperbolicity_new_improved(data, g);
    end = getTimeSeconds();

    fprintf(stdout, "%7.1f%17ld%11.4f", delta, nt, end - begin);
    fprintf(out, "%7.1f%17ld%11.4f", delta, nt, end - begin);
    oldnt = nt;
    nt = 0;
    fprintf(stdout, "%11.4f", t);
    fprintf(out, "%11.4f", t);
    fflush(stdout);
    fflush(out);

    begin = getTimeSeconds();
    deltacheck = hyperbolicity_new(data, g);
    end = getTimeSeconds();

    if (abs(deltacheck - delta) > 1E-5 || oldnt != nt) {
        fprintf(stdout, "   MISTAKE");
        fprintf(out, "   MISTAKE");
    } else {
        fprintf(stdout, "%11.4f", end - begin);
        fprintf(out, "%11.4f", end - begin);
    }
    fflush(stdout);
    fflush(out);
    nt = 0;
    begin = getTimeSeconds();
    deltacheck = hyperbolicity_old(data, g);
    end = getTimeSeconds();

    if (abs(deltacheck - delta) > 1E-5 ) {
        fprintf(stdout, "          MISTAKE          \n");
        fprintf(out, "          MISTAKE          \n");
    } else {
        fprintf(stdout, "%17ld", nt);
        fprintf(out, "%17ld", nt);
        fprintf(stdout, "%11.4f\n", end - begin);
        fprintf(out, "%11.4f\n", end - begin);
    }

    fflush(stdout);
    fflush(out);

    fclose(out);
    graph_free(g);
    data_free(data);
    return (0);
}
